#include "sv_server_filter_VolumeFilter.h"
#include "volume.h"

JNIEXPORT jint JNICALL Java_sv_server_filter_VolumeFilter_executeFilter
               (JNIEnv *env, jclass clazz, 
                jstring prefix, jbyteArray arr) {


#define COLOR_CHANNELS   3

  jboolean isCopy;
  const char *name = (*env)->GetStringUTFChars(env, prefix, &isCopy);

  vpContext *vpc; 
  int clvolume_fd;
  float density_ramp[DENSITY_MAX+1];
  float gradient_ramp[GRADIENT_MAX+1];


  float color_table[(NORMAL_MAX+1) * COLOR_CHANNELS];
  float weight_table[DENSITY_MAX+1];

  static unsigned char vimage[IMAGE_WIDTH][IMAGE_HEIGHT][3];


  int n, sz, i, j;
  char filename[512];


  jbyte *data_arr = (*env)->GetByteArrayElements(env, arr, 0);
  jfieldID fid;
  double ray_red, ray_green, ray_blue;
  double rotx, roty;

  fid = (*env)->GetStaticFieldID(env, clazz, "rayColorRed", "D");
  if (fid == 0)
    return;
  ray_red = (*env)->GetStaticDoubleField(env, clazz, fid); 

  fid = (*env)->GetStaticFieldID(env, clazz, "rayColorGreen", "D");
  if (fid == 0)
    return;
  ray_green = (*env)->GetStaticDoubleField(env, clazz, fid); 

  fid = (*env)->GetStaticFieldID(env, clazz, "rayColorBlue", "D");
  if (fid == 0)
    return;
  ray_blue = (*env)->GetStaticDoubleField(env, clazz, fid); 

  fid = (*env)->GetStaticFieldID(env, clazz, "rotx", "D");
  if (fid == 0)
    return;
  rotx = (*env)->GetStaticDoubleField(env, clazz, fid); 

  fid = (*env)->GetStaticFieldID(env, clazz, "roty", "D");
  if (fid == 0)
    return;
  roty = (*env)->GetStaticDoubleField(env, clazz, fid); 

  vpc = vpCreateContext();

  sprintf(filename, "%s.cv", name);
  if ((clvolume_fd = open(filename, 0)) < 0) {
     fprintf(stderr, "could not open %s\n", filename);
     return -1;
  }
  if (vpLoadClassifiedVolume(vpc, clvolume_fd) != VP_OK) {
     fprintf(stderr, "VolPack error: %s\n",
	    vpGetErrorString(vpGetError(vpc)));
     fprintf(stderr, "could not load the volume from file %s\n", filename);
    
     return -1;
  }
  close(clvolume_fd);


  vpSetLookupShader(vpc, COLOR_CHANNELS, 1, 
                NORMAL_FIELD, color_table, sizeof(color_table), 
                DENSITY_FIELD, weight_table, sizeof(weight_table));

  vpSetMaterial(vpc, VP_MATERIAL0, VP_AMBIENT, VP_BOTH_SIDES,
                      0.18, 0.18, 0.18);
  vpSetMaterial(vpc, VP_MATERIAL0, VP_DIFFUSE, VP_BOTH_SIDES,
	              0.35, 0.35, 0.35);
  vpSetMaterial(vpc, VP_MATERIAL0, VP_SPECULAR, VP_BOTH_SIDES,
                      0.39, 0.39, 0.39);
  vpSetMaterial(vpc, VP_MATERIAL0, VP_SHINYNESS, VP_BOTH_SIDES,
                     10.0,0.0,0.0);
  vpSetLight(vpc, VP_LIGHT0, VP_DIRECTION, 0.3, 0.3, 1.0);
  vpSetLight(vpc, VP_LIGHT0, VP_COLOR, ray_red, ray_green, ray_blue);
  vpEnable(vpc, VP_LIGHT0, 1);
  vpSetDepthCueing(vpc, 1.4, 1.5);
  vpEnable(vpc, VP_DEPTH_CUE, 1);


  vpSeti(vpc, VP_CONCAT_MODE, VP_CONCAT_LEFT);
  vpRotate(vpc, VP_Y_AXIS, roty);
  vpRotate(vpc, VP_X_AXIS, rotx);
  vpCurrentMatrix(vpc, VP_PROJECT);
  vpIdentityMatrix(vpc);
  vpWindow(vpc, VP_PARALLEL, -0.5, 0.5, -0.5, 0.5, -0.5, 0.5);
  vpCurrentMatrix(vpc, VP_MODEL);

  vpSetImage(vpc, (unsigned char *)vimage, IMAGE_WIDTH, IMAGE_HEIGHT,
	       3*IMAGE_WIDTH, VP_RGB);

  vpSetd(vpc, VP_MAX_RAY_OPACITY, 0.95);

  if (vpShadeTable(vpc) != VP_OK) {
     fprintf(stderr, "VolPack error: %s\n", vpGetErrorString(vpGetError(vpc)));
     return -1;
  }

  if (vpRenderClassifiedVolume(vpc) != VP_OK) {
     fprintf(stderr, "VolPack error: %s\n", vpGetErrorString(vpGetError(vpc)));
     return -1;
  }

  for (i=0; i<IMAGE_HEIGHT; i++) 
    for (j=0; j<IMAGE_WIDTH; j++) {
       n = 3 * (i*IMAGE_WIDTH + j);
       data_arr[n] = (jbyte)vimage[i][j][0];
       data_arr[n+1] = (jbyte)vimage[i][j][1];
       data_arr[n+2] = (jbyte)vimage[i][j][2];
    }

  if (isCopy == JNI_TRUE)
    (*env)->ReleaseByteArrayElements(env, arr, data_arr, 0);

  (*env)->ReleaseStringUTFChars(env, prefix, name); 

  return 0;
}
